import threading
import time
import logging
from gql import gql
from datetime import datetime
from dateutil.parser import isoparse

from .data import Issue
from .data import Repository


def _make_issue(gh_data, is_pr=False):
    repo = Repository(
        owner=gh_data["repository"]["owner"]["login"],
        name=gh_data["repository"]["name"],
    )
    if gh_data["author"] is None:
        # https://github.com/ghost
        author = "ghost"
    else:
        author = gh_data["author"]["login"]
    return Issue(
        repo=repo,
        author=author,
        created_at=isoparse(gh_data["createdAt"]),
        updated_at=isoparse(gh_data["updatedAt"]),
        number=int(gh_data["number"]),
        title=gh_data["title"],
        url=gh_data["url"],
        is_read=bool(gh_data["isReadByViewer"]),
        is_pr=is_pr,
    )


FRAGMENT_ISSUE = """
fragment issueFields on Issue {
    author {
        login
    }
    createdAt
    number
    title
    updatedAt
    url
    isReadByViewer
    repository {
        name
        owner {
            login
        }
    }
}
"""
FRAGMENT_PR = """
fragment prFields on PullRequest {
    author {
        login
    }
    createdAt
    number
    title
    updatedAt
    url
    isReadByViewer
    repository {
        name
        owner {
            login
        }
    }
}
"""


class IssueQuery:

    def __init__(self, *, first=100, after=None, states=("OPEN",)):
        self.first = first
        self.after = after
        self.states = states

    def __str__(self):
        parts = [f"issues(first: {self.first}"]
        if self.after:
            parts.append(f', after: "{self.after}"')
        if self.states:
            parts.append(", states: [")
            parts.append(",".join(self.states))
            parts.append("]")
        parts.append(
            ") { nodes { ...issueFields } pageInfo { endCursor hasNextPage } }"
        )
        return "".join(parts)


class PRQuery:

    def __init__(self, *, first=100, after=None, states=("OPEN",)):
        self.first = first
        self.after = after
        self.states = states

    def __str__(self):
        parts = [f"pullRequests(first: {self.first}"]
        if self.after:
            parts.append(f', after: "{self.after}"')
        if self.states:
            parts.append(", states: [")
            parts.append(",".join(self.states))
            parts.append("]")
        parts.append(") { nodes { ...prFields } pageInfo { endCursor hasNextPage } }")
        return "".join(parts)


class IssueLoader:

    def __init__(self, gql_client, repos, cache, progress_callback):
        self._client = gql_client
        self._repos = tuple(repos)
        self._cache = cache
        self._lock = threading.Lock()
        self._logger = logging.getLogger("IssueLoader")
        self._thread = threading.Thread(daemon=True, target=self._run)
        self._progress_callback = progress_callback
        self._thread.start()

    def _run(self):
        try:
            self._load_all_issues(
                repos_per_query=10, progress_callback=self._progress_callback
            )
        except:
            self._logger.exception("Exception in IssueLoader thread")
        while True:
            time.sleep(15)
            try:
                # Uses search API to get updated issues and PRs
                self._update_all_issues()
            except:
                self._logger.exception("Exception in IssueLoader thread")

    def _load_all_issues(self, repos_per_query, progress_callback=None):
        num_repos_at_start = len(self._repos)

        # These dicts indicate if repos have more issues or PRs to query
        issue_page_info = {}
        pr_page_info = {}
        for r in self._repos:
            # Use "None" to mean we haven't queried anything yet
            issue_page_info[r] = {"endCursor": None}
            pr_page_info[r] = {"endCursor": None}

        # Outer loop runs until it finishes exploring all issues and PRs
        # on all repos
        next_queries = []
        issue_count = 0
        pr_count = 0
        while issue_page_info or pr_page_info:
            # This loop looks for repos that still need to be explored
            for r in self._repos:
                issue_query = None
                pr_query = None
                if r in issue_page_info:
                    ipi = issue_page_info[r]
                    issue_query = IssueQuery(after=ipi["endCursor"])
                if r in pr_page_info:
                    prpi = pr_page_info[r]
                    pr_query = PRQuery(after=prpi["endCursor"])

                if issue_query or pr_query:
                    repo_query = (
                        f'r{id(r)}: repository(owner: "{r.owner}", name: "{r.name}")'
                    )
                    repo_query += "{"
                    if issue_query:
                        repo_query += str(issue_query)
                    if pr_query:
                        repo_query += str(pr_query)
                    repo_query += "}"
                    next_queries.append(repo_query)

                if len(next_queries) == repos_per_query:
                    # Found enough repos, ditch the for loop
                    break

            # Execute the next query
            joined_queries = "\n".join(next_queries)
            query_str = f"""
                query {{
                    {joined_queries}
                }}
                """
            next_queries = []
            # It's an error to include unused fragments,
            # so only include a fragment if it's used.
            if issue_page_info:
                query_str += FRAGMENT_ISSUE
            if pr_page_info:
                query_str += FRAGMENT_PR
            query = gql(query_str)
            result = self._client.execute(query)

            # Figure out what repos the query included,
            # and add the issues and PRs to the cache
            for r in self._repos:
                key = f"r{id(r)}"
                if key not in result:
                    # Query didn't include this repo
                    continue
                repo_result = result[key]
                if "issues" in repo_result:
                    for issue in repo_result["issues"]["nodes"]:
                        self._cache.insert(_make_issue(issue))
                        issue_count += 1
                    if repo_result["issues"]["pageInfo"]["hasNextPage"]:
                        issue_page_info[r] = repo_result["issues"]["pageInfo"]
                    else:
                        del issue_page_info[r]
                if "pullRequests" in repo_result:
                    for pr in repo_result["pullRequests"]["nodes"]:
                        self._cache.insert(_make_issue(pr, is_pr=True))
                        pr_count += 1
                    if repo_result["pullRequests"]["pageInfo"]["hasNextPage"]:
                        pr_page_info[r] = repo_result["pullRequests"]["pageInfo"]
                    else:
                        del pr_page_info[r]
            if progress_callback:
                i_max = pr_max = num_repos_at_start
                progress_callback(
                    (i_max - len(issue_page_info) + pr_max - len(pr_page_info))
                    / (i_max + pr_max)
                )
        self._logger.info(f"Loaded {issue_count} issues and {pr_count} PRs")

    def _update_all_issues(self):

        updated_time = self._cache.newest_update_time().isoformat()

        def make_query(extra):
            gh_search = f"{extra} is:open updated:>{updated_time} "
            gh_search += " ".join([f"repo:{r.owner}/{r.name}" for r in self._repos])
            query_parts = ["{"]
            query_parts.append(f'search(first: 1, query: "{gh_search}", type: ISSUE) ')
            query_parts.append("{ nodes {...issueFields ...prFields} }")
            query_parts.append("}")
            query_parts.append(FRAGMENT_ISSUE)
            query_parts.append(FRAGMENT_PR)
            return " ".join(query_parts)

        # TODO use pagination for unlikely case of more than 100 updated issues
        # Must query for issues and PRs separately
        # https://github.com/orgs/community/discussions/149046
        issues = self._client.execute(gql(make_query("is:issue")))
        for i in issues["search"]["nodes"]:
            self._cache.insert(_make_issue(i))

        prs = self._client.execute(gql(make_query("is:pr")))
        for pr in prs["search"]["nodes"]:
            self._cache.insert(_make_issue(pr, is_pr=True))



import abc
import threading
import time

from gql import gql

import requests

from yaml import safe_load as load_yaml

from .data import Repository


class RepoLoader(abc.ABC):

    def __init__(self):
        self._done_callback = None
        self._repos = None
        self._thread = threading.Thread(target=self._load_repos, daemon=True)

    def begin_loading(self, done_callback):
        self._done_callback = done_callback
        self._thread.start()

    @abc.abstractmethod
    def load_repos(self) -> tuple[Repository]: ...

    def _load_repos(self):
        self._repos = self.load_repos()
        self._done_callback(self._repos)
        self._done_callback = None

    def repos(self) -> tuple[Repository]:
        return self._repos

    def cleanup(self):
        self._thread.join()
        self._thread = None


class SequentialRepoLoaders(RepoLoader):
    """Invokes multiple repo loaders in a chain."""

    def __init__(self, repo_loaders, *args, **kwargs):
        self._loaders = repo_loaders
        super().__init__(*args, **kwargs)

    def load_repos(self):
        repos = []
        for loader in self._loaders:
            repos.extend(loader.load_repos())
        repos = tuple(set(repos))
        return repos


class CurrentUserRepoLoader(RepoLoader):

    def __init__(self, gql_client, *args, **kwargs):
        self._client = gql_client
        super().__init__(*args, **kwargs)

    def load_repos(self):
        repos = []

        def _query(after=""):
            query = gql(
                """
                query($after: String!) {
                    viewer {
                        repositories(after: $after, first: 100, visibility: PUBLIC, affiliations: [OWNER], isArchived: false) {
                            nodes {
                                nameWithOwner
                            }
                            pageInfo {
                                endCursor
                                hasNextPage
                            }
                        }
                    }
                }
                """
            )
            result = self._client.execute(query, variable_values={"after": after})
            return result

        q = None
        while q is None or q["viewer"]["repositories"]["pageInfo"]["hasNextPage"]:
            if q is None:
                q = _query("")
            else:
                q = _query(q["viewer"]["repositories"]["pageInfo"]["endCursor"])
            for r in q["viewer"]["repositories"]["nodes"]:
                owner, name = r["nameWithOwner"].split("/")
                repos.append(Repository(name=name, owner=owner))
        return tuple(repos)


class OrgRepoLoader(RepoLoader):

    def __init__(self, organization, gql_client, *args, **kwargs):
        self._client = gql_client
        self.organization = organization
        super().__init__(*args, **kwargs)

    def load_repos(self):
        repos = []

        def _query(after=""):
            query = gql(
                """
                query($after: String!, $organization: String!) {
                    organization(login: $organization) {
                        repositories(after: $after, first: 100, visibility: PUBLIC, isArchived: false) {
                            nodes {
                                nameWithOwner
                            }
                            pageInfo {
                                endCursor
                                hasNextPage
                            }
                        }
                    }
                }
                """
            )
            result = self._client.execute(
                query,
                variable_values={"after": after, "organization": self.organization},
            )
            return result

        q = None
        while q is None or q["organization"]["repositories"]["pageInfo"]["hasNextPage"]:
            if q is None:
                q = _query("")
            else:
                q = _query(q["organization"]["repositories"]["pageInfo"]["endCursor"])
            for r in q["organization"]["repositories"]["nodes"]:
                owner, name = r["nameWithOwner"].split("/")
                repos.append(Repository(name=name, owner=owner))
        return tuple(repos)


class FileRepoLoader(RepoLoader):

    def __init__(self, filepath, *args, **kwargs):
        self.filepath = filepath
        super().__init__(*args, **kwargs)

    def load_repos(self):
        repos = []

        with self.filepath.open() as f:
            while line := f.readline():
                print(line)
                # Strip comments
                if "#" in line:
                    line = line[: line.find("#")]
                # strip whitespace
                line = line.strip()
                if not line:
                    # ingore blank lines
                    continue
                if line.count("/") != 1:
                    raise RuntimeError(
                        "Incorrect format: {line}. Use one owner/name per line."
                    )
                owner, name = line.split("/")
                print(line)
                repos.append(Repository(name=name, owner=owner))
        return tuple(repos)


class VcsRepoLoader(RepoLoader):

    def __init__(self, url, *args, **kwargs):
        self.url = url
        super().__init__(*args, **kwargs)

    def load_repos(self):
        repos = []

        r = requests.get(self.url)
        if r.status_code != 200:
            raise RuntimeError(f"TODO Handle VCS Repose download failure {r}")

        repos_file = load_yaml(r.text)
        if "repositories" not in repos_file:
            raise RuntimeError(f"TODO handle invalid repos file {repos_file}")

        for repo_data in repos_file["repositories"].values():
            if "url" not in repo_data:
                raise RuntimeError(f"TODO handle invalid repos file {repos_file}")
            url = repo_data["url"]
            if not url.startswith("https://github.com/"):
                continue
            url = url[len("https://github.com/") :]
            if url.endswith(".git"):
                url = url[: -len(".git")]
            if url.count("/") != 1:
                raise RuntimeError("Unable to parse url: {repo_data['url']}")
            owner, name = url.split("/")
            repos.append(Repository(name=name, owner=owner))

        return tuple(repos)



import lark
import pathlib

from treadi.data import Issue

GRAMMAR = (pathlib.Path(__file__).parent.resolve() / "filter.lark").read_text()
PARSER = lark.Lark(GRAMMAR, parser="lalr")


def require_pr(issue: Issue):
    """Return True if the issue is a PR."""
    return issue.is_pr


def require_issue(issue: Issue):
    """Return True if the issue is an issue."""
    return not issue.is_pr


class RequireAuthor:

    def __init__(self, author):
        self._author = author

    def __call__(self, issue: Issue):
        """Return True if the issue is authored by this author."""
        return self._author.lower() == issue.author.lower()


class RequireRepo:

    def __init__(self, owner, name):
        self._owner = owner
        self._name = name

    def __call__(self, issue: Issue):
        """Return True if the issue belongs to the given repo."""
        return (
            self._owner.lower() == issue.repo.owner.lower()
            and self._name.lower() == issue.repo.name.lower()
        )


class InvertRequirement:

    def __init__(self, requirement):
        self._requirement = requirement

    def __call__(self, issue: Issue):
        """Return the opposite of what the requirement returns."""
        return not self._requirement(issue)


class RequireAll:

    def __init__(self, requirements):
        self._requirements = tuple(requirements)

    def __call__(self, issue: Issue):
        """Return True iff all requirements are met."""
        for requirement in self._requirements:
            if not requirement(issue):
                return False
        return True


class FilterTransformer(lark.Transformer):
    """Turn a lark parse tree into a callable that filters issues."""

    def start(self, args):
        return RequireAll(args)

    def statement(self, args):
        return args[0]

    def maybe_negated_statement(self, args):
        if args[0] == "-":
            return InvertRequirement(args[1])
        return args[0]

    def type_stmt(self, args):
        if args[0] == "pr":
            return require_pr
        return require_issue

    def author_stmt(self, args):
        return RequireAuthor(args[0].value)

    def assignee_stmt(self, args):
        raise NotImplementedError

    def is_stmt(self, args):
        if args[0] == "pr":
            return require_pr
        if args[0] == "issue":
            return require_issue
        raise NotImplementedError

    def review_stmt(self, args):
        raise NotImplementedError

    def repo_stmt(self, args):
        return RequireRepo(args[0].value, args[1].value)


def parse(filter: str):
    tree = PARSER.parse(filter)
    return FilterTransformer().transform(tree)



from treadi.data import Issue
from treadi.data import Repository
import treadi.filter as tf

REPO_ROS_ROSDITRO = Repository(
    owner="ros",
    name="rosdistro",
)
REPO_ROS2_ROS2 = Repository(
    owner="ros2",
    name="ros2",
)
ISSUE_ROS_ROSDISTRO = Issue(
    repo=REPO_ROS_ROSDITRO,
    author="sloretz",
    created_at=None,
    updated_at=None,
    number=42,
    title="fix stuff",
    url="",
    is_read=False,
    is_pr=False,
)
PR_ROS_ROSDISTRO = Issue(
    repo=REPO_ROS_ROSDITRO,
    author="ghost",
    created_at=None,
    updated_at=None,
    number=1,
    title="break stuff",
    url="",
    is_read=False,
    is_pr=True,
)


def test_require_pr():
    assert tf.require_pr(PR_ROS_ROSDISTRO)
    assert not tf.require_pr(ISSUE_ROS_ROSDISTRO)


def test_require_issue():
    assert tf.require_issue(ISSUE_ROS_ROSDISTRO)
    assert not tf.require_issue(PR_ROS_ROSDISTRO)


def test_require_author():
    assert tf.RequireAuthor("sloretz")(ISSUE_ROS_ROSDISTRO)
    assert not tf.RequireAuthor("bob")(ISSUE_ROS_ROSDISTRO)


def test_require_repo():
    assert tf.RequireRepo("ros", "rosdistro")(ISSUE_ROS_ROSDISTRO)
    assert not tf.RequireRepo("ros2", "ros2")(ISSUE_ROS_ROSDISTRO)


def test_invert_requirement():
    assert not tf.InvertRequirement(tf.require_pr)(PR_ROS_ROSDISTRO)
    assert tf.InvertRequirement(tf.require_pr)(ISSUE_ROS_ROSDISTRO)


def test_require_all():
    assert tf.RequireAll([tf.RequireAuthor("sloretz"), tf.require_issue])(
        ISSUE_ROS_ROSDISTRO
    )
    assert not tf.RequireAll([tf.RequireAuthor("ghost"), tf.require_issue])(
        PR_ROS_ROSDISTRO
    )
    assert not tf.RequireAll([tf.RequireAuthor("sloretz"), tf.require_issue])(
        PR_ROS_ROSDISTRO
    )



import lark

import pytest

from treadi.filter import GRAMMAR


@pytest.fixture
def parser():
    return lark.Lark(GRAMMAR, parser="lalr", debug=True, strict=True)


def test_negation(parser):
    parser.parse("-type:issue")
    parser.parse("-author:slorretz")
    parser.parse("-repo:ros/rosdistro")


def test_multiple(parser):
    parser.parse("-type:issue author:sloretz -repo:ros/rodistro")


def test_type(parser):
    parser.parse("type:issue")
    parser.parse("type:pr")
    with pytest.raises(lark.UnexpectedToken):
        parser.parse("type:foo")
    with pytest.raises(lark.UnexpectedToken):
        parser.parse("type: issue")


def test_author(parser):
    parser.parse("author:sloretz")
    with pytest.raises(lark.UnexpectedToken):
        parser.parse("author:-sloretz")


def test_assignee(parser):
    parser.parse("assignee:sloretz")
    with pytest.raises(lark.UnexpectedCharacters):
        parser.parse("assignee:*sloretz")


def test_is(parser):
    parser.parse("is:draft")
    with pytest.raises(lark.UnexpectedToken):
        parser.parse("is:candy")


def test_review(parser):
    parser.parse("review:none")
    parser.parse("review:approved")
    parser.parse("review:changes_requested")
    with pytest.raises(lark.UnexpectedToken):
        parser.parse("review:candy")


def test_repo(parser):
    parser.parse("repo:sloretz/TreadI")
    with pytest.raises(lark.UnexpectedToken):
        parser.parse("repo:sloretz")



import copy
import random

from dateutil.parser import isoparse

from treadi.data import Issue
from treadi.data import Repository
from treadi.issue_cache import IssueCache


def random_string(*, len=5):
    alphabet = "abcdefghijklmnopqrstuvwxyz"
    numbers = "0123456789"
    symbols = "_-"
    return "".join(
        random.sample(alphabet + alphabet.upper() + numbers + symbols, k=len)
    )


def rand_repo():
    return Repository(owner=random_string(), name=random_string())


def rand_issue(*, updated_at, repo=None, is_read=False):
    if repo is None:
        repo = rand_repo()
    return Issue(
        repo=repo,
        author=random_string(),
        created_at=isoparse("2006-07-04T15:00:00Z"),
        updated_at=isoparse(updated_at),
        number=random.randint(1, 9999),
        title=random_string(),
        url=random_string(),
        is_read=is_read,
    )


def test_cache_sorted():
    cache = IssueCache()
    first = rand_issue(updated_at="2006-07-04T15:00:00Z")
    second = rand_issue(updated_at="2006-07-04T16:00:00Z")
    third = rand_issue(updated_at="2006-07-04T17:00:00Z")
    fourth = rand_issue(updated_at="2006-07-04T18:00:00Z")
    fifth = rand_issue(updated_at="2006-07-04T19:00:00Z")
    sixth = rand_issue(updated_at="2006-07-04T20:00:00Z")
    issues_to_add = [
        first,
        second,
        third,
        fourth,
        fifth,
        sixth,
    ]
    random.shuffle(issues_to_add)
    for i in issues_to_add:
        cache.insert(i)
    assert [sixth, fifth, fourth] == cache.most_recent_issues(3)


def test_cache_dismissed():
    cache = IssueCache()
    first = rand_issue(updated_at="2006-07-04T15:00:00Z")
    second = rand_issue(updated_at="2006-07-04T16:00:00Z")
    third = rand_issue(updated_at="2006-07-04T17:00:00Z")
    cache.insert(first)
    cache.insert(second)
    cache.insert(third)
    cache.dismiss(second)
    assert [third, first] == cache.most_recent_issues(3)
    cache.insert(second)
    assert [third, first] == cache.most_recent_issues(3)


def test_cache_update_dismissed():
    cache = IssueCache()
    issue = rand_issue(updated_at="2006-07-04T15:00:00Z")
    cache.insert(issue)
    cache.dismiss(issue)
    assert [] == cache.most_recent_issues(1)
    issue = copy.deepcopy(issue)
    issue.updated_at = isoparse("2006-07-04T16:00:00Z")
    cache.insert(issue)
    assert [issue] == cache.most_recent_issues(1)


def test_cache_update_dismissed_is_read():
    cache = IssueCache()
    issue = rand_issue(updated_at="2006-07-04T15:00:00Z")
    cache.insert(issue)
    cache.dismiss(issue)
    assert [] == cache.most_recent_issues(1)
    issue = copy.deepcopy(issue)
    issue.updated_at = isoparse("2006-07-04T16:00:00Z")
    issue.is_read = True
    cache.insert(issue)
    assert [] == cache.most_recent_issues(1)



from dataclasses import dataclass

from enum import Enum
import requests
import urllib

from . import CLIENT_ID

import keyring


SERVICE = "TreadI"
USERNAME = "GithubRefreshToken"


def get_refresh_token():
    refresh_token = keyring.get_password(SERVICE, USERNAME)
    if refresh_token is not None:
        keyring.delete_password(SERVICE, USERNAME)
    return refresh_token


def store_refresh_token(refresh_token):
    keyring.set_password(SERVICE, USERNAME, refresh_token)


@dataclass
class DeviceFlow:

    device_code: str
    user_code: str
    verification_uri: str
    interval: int
    expires_in: int


def start_device_flow(*, client_id=CLIENT_ID):
    url = "https://github.com/login/device/code"
    r = requests.post(url, data={"client_id": client_id})
    if r.status_code != 200:
        raise RuntimeError(f"TODO Handle device code request failure {r}")
    response = urllib.parse.parse_qs(r.text)

    return DeviceFlow(
        device_code=response["device_code"][0],
        user_code=response["user_code"][0],
        verification_uri=response["verification_uri"][0],
        interval=int(response["interval"][0]),
        expires_in=int(response["expires_in"][0]),
    )


class Status(Enum):
    # The user gave TreadI the access it needs
    ACCESS_GRANTED = 1
    # Waiting for the user to make a decision
    AUTHORIZATION_PENDING = 10
    # User took too long. The process needs to be started over
    EXPIRED_TOKEN = 12
    # User denied TreadI access
    ACCESS_DENIED = 13
    # Refresh token is invalid
    INCORRECT_CLIENT_CREDENTIALS = 51
    # Something else that TreadI doesn't handle specifically
    OTHER_ERROR = 100
    # Not a github error, this means there's no cached refresh token
    NO_TOKEN = 1001


@dataclass
class TokenResponse:
    status: Status
    access_token: str = None
    refresh_token: str = None


def ask_for_token(device_flow, *, client_id=CLIENT_ID):
    data = {
        "client_id": client_id,
        "device_code": device_flow.device_code,
        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
    }
    r = requests.post("https://github.com/login/oauth/access_token", data=data)
    if r.status_code != 200:
        return TokenResponse(status=Status.OTHER_ERROR)
    response = urllib.parse.parse_qs(r.text)
    print(response)
    if "interval" in response:
        device_flow.interval = int(response["interval"][0])
    if "error" in response.keys():
        match response["error"][0]:
            case "slow_down" | "authorization_pending":
                return TokenResponse(status=Status.AUTHORIZATION_PENDING)
            case "access_denied":
                return TokenResponse(status=Status.ACCESS_DENIED)
            case _:
                return TokenResponse(status=Status.OTHER_ERROR)
    return TokenResponse(
        status=Status.ACCESS_GRANTED,
        access_token=response["access_token"][0],
        refresh_token=response["refresh_token"][0],
    )


def refresh_access_token(refresh_token, *, client_id=CLIENT_ID):
    data = {
        "client_id": client_id,
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
    }
    r = requests.post("https://github.com/login/oauth/access_token", data=data)
    if r.status_code != 200:
        return TokenResponse(status=Status.OTHER_ERROR)
    response = urllib.parse.parse_qs(r.text)
    if "error" in response.keys():
        match response["error"][0]:
            case "incorrect_client_credentials":
                return TokenResponse(status=Status.INCORRECT_CLIENT_CREDENTIALS)
            case _:
                return TokenResponse(status=Status.OTHER_ERROR)
    return TokenResponse(
        status=Status.ACCESS_GRANTED,
        access_token=response["access_token"][0],
        refresh_token=response["refresh_token"][0],
    )


def cycle_cached_token():
    rt = get_refresh_token()
    if rt is None:
        return TokenResponse(status=Status.NO_TOKEN)
    response = refresh_access_token(rt)
    if response.status == Status.ACCESS_GRANTED:
        store_refresh_token(response.refresh_token)
    return response



import kivy

kivy.require("2.3.1")

from kivy.animation import Animation
from kivy.app import App
from kivy.clock import Clock
from kivy.config import Config

Config.set("graphics", "resizable", False)
Config.set("input", "mouse", "mouse,disable_multitouch")
from kivy.core.window import Window

Window.size = (500, 558)
from kivy.properties import ColorProperty
from kivy.properties import NumericProperty
from kivy.properties import ObjectProperty

from kivy.uix.behaviors import ButtonBehavior
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.stacklayout import StackLayout
from kivy.uix.widget import Widget
from kivy.uix.screenmanager import ScreenManager, Screen

import urllib
import webbrowser
import requests
import threading
import time
import logging
import pathlib

from gql import gql, Client
from gql.transport.requests import RequestsHTTPTransport

from . import auth
from .data import Issue
from .data import is_same_issue
from .issue_cache import IssueCache
from .issue_loader import IssueLoader
from .repo_loader import CurrentUserRepoLoader
from .repo_loader import OrgRepoLoader
from .repo_loader import FileRepoLoader
from .repo_loader import SequentialRepoLoaders
from .repo_loader import VcsRepoLoader
from .filter import parse as parse_filter


USERNAME = None
REPOS = None
ISSUES = None


def make_gql_client(access_token):
    transport = RequestsHTTPTransport(
        url="https://api.github.com/graphql",
        headers={
            "Authorization": f"bearer {access_token}",
        },
        verify=True,
        retries=3,
    )
    schema_path = pathlib.Path(__file__).parent.resolve() / "schema.docs.graphql"
    return Client(transport=transport, schema=schema_path.read_text())


class IssueWidget(ButtonBehavior, BoxLayout):

    color = ColorProperty(defaultvalue=[0.6, 0.6, 0.6, 1])

    issue = ObjectProperty(
        Issue(),
        rebind=True,
    )

    def __init__(self, issue, dismiss_callback, **kwargs):
        self.issue = issue
        self.dismiss_callback = dismiss_callback
        super().__init__(**kwargs)

    def on_press(self):
        if self.last_touch.button == "left":
            self.color = [0.6, 0.6, 0.8, 1]

    def on_release(self):
        # Oh yeah, the whole point is to open the issue
        if self.last_touch.button == "left":
            webbrowser.open(self.issue.url)

    def do_dismiss_callback(self):
        # Only dismiss once
        if self.dismiss_callback is not None:
            d = self.dismiss_callback
            self.dismiss_callback = None
            d(self)


class IssueScreen(Screen):

    def __init__(self, *args, **kwargs):
        self._filter = None
        self._logger = logging.getLogger("IssueLoader")
        super().__init__(*args, **kwargs)

    def validate_filter(self):
        # Called when on_validate_text is called on filter TextInput
        filter_str = self.ids.filter.text
        if filter_str.strip() == "":
            self._filter = None
            self.ids.filter.background_color = (1, 1, 1, 1)
        else:
            try:
                self._filter = parse_filter(filter_str.strip())
                self.ids.filter.background_color = (0.9, 1, 0.9, 1)
            except:
                self._logger.exception("Cannot parse filter")
                self._filter = None
                self.ids.filter.background_color = (1, 0.5, 0.5, 1)

        self._refresh_issues()

    def on_text_changing(self):
        # called hen on_text is called on filter TextInput
        # Filters don't apply until the user hits enter, so display as yellow
        self.ids.filter.background_color = (1, 1, 0.9, 1)

    def on_pre_enter(self):
        issue_cache = App.get_running_app().issue_cache
        self._refresh_issues()

    def _refresh_issues(self):
        # Clear and re-add issues
        self.ids.stack.clear_widgets()
        issue_cache = App.get_running_app().issue_cache
        issues = issue_cache.most_recent_issues(n=5, filter=self._filter)
        for i in issues:
            self._add_issue(i)

    def _add_issue(self, issue):
        self.ids.stack.add_widget(IssueWidget(issue, self.dismiss))

    def dismiss(self, issue_widget):
        issue = issue_widget.issue
        issue_cache = App.get_running_app().issue_cache

        issue_cache.dismiss(issue)

        child_issues = []
        for child in self.ids.stack.children:
            if hasattr(child, "issue"):
                child_issues.append(child.issue)

        # aim for 1 additional issue, because one is being dismissed
        consider_issues = issue_cache.most_recent_issues(
            n=1 + len(child_issues), filter=self._filter
        )
        for itc in consider_issues:
            displaying_issue = False
            for chi in child_issues:
                if is_same_issue(itc, chi):
                    displaying_issue = True
                    break
            if not displaying_issue:
                self._add_issue(itc)
                break

        # Animate the dismissed widget shrinking, so new issue reveals from below
        anim = Animation(
            size_hint_y=0, opacity=0, duration=0.125, transition="out_cubic"
        )
        anim.bind(on_complete=lambda *args: self.ids.stack.remove_widget(issue_widget))
        anim.start(issue_widget)


class RepoPickerScreen(Screen):

    def use_all_user_repos(self):
        self.manager.switch_to(
            RepoLoadingScreen(CurrentUserRepoLoader(App.get_running_app().gql_client))
        )

    def use_all_gazebo_repos(self):
        self.manager.switch_to(
            RepoLoadingScreen(
                SequentialRepoLoaders(
                    repo_loaders=[
                        OrgRepoLoader("gazebosim", App.get_running_app().gql_client),
                        OrgRepoLoader(
                            "gazebo-tooling", App.get_running_app().gql_client
                        ),
                        OrgRepoLoader(
                            "gazebo-release", App.get_running_app().gql_client
                        ),
                    ],
                )
            )
        )

    def use_all_rmf_repos(self):
        self.manager.switch_to(
            RepoLoadingScreen(
                OrgRepoLoader("open-rmf", App.get_running_app().gql_client)
            )
        )

    def use_all_infra_repos(self):
        self.manager.switch_to(
            RepoLoadingScreen(
                VcsRepoLoader(
                    "https://raw.githubusercontent.com/ros-infrastructure/ci/refs/heads/main/ros-infrastructure.repos"
                )
            )
        )

    def use_all_ros_repos(self):
        self.manager.switch_to(
            RepoLoadingScreen(
                FileRepoLoader(
                    pathlib.Path(__file__).parent.resolve() / "ros_pmc_repos.txt"
                )
            )
        )


class RepoLoadingScreen(Screen):

    def __init__(self, loader, **kwargs):
        self._loader = loader
        self._loader.begin_loading(self.switch_to_issue_loading)
        super().__init__(**kwargs)

    def switch_to_issue_loading(self, repos):

        def _switch(dt):
            self.manager.switch_to(IssueLoadingScreen(repos))

        Clock.schedule_once(lambda dt: self._loader.cleanup())
        Clock.schedule_once(_switch)


class IssueLoadingScreen(Screen):

    progress = NumericProperty(0.0)

    def __init__(self, repos, **kwargs):
        App.get_running_app().issue_loader = IssueLoader(
            App.get_running_app().gql_client,
            repos,
            App.get_running_app().issue_cache,
            self.update_progress,
        )
        super().__init__(**kwargs)

    def update_progress(self, progress):
        self.progress = progress * 100
        if progress >= 1.0:
            Clock.schedule_once(lambda dt: self.switch_to_issues())

    def switch_to_issues(self):
        # Must only be called on main thread
        self.manager.transition.direction = "left"
        self.manager.current = "issues"


class LoginScreen(Screen):

    device_flow = ObjectProperty(
        auth.DeviceFlow(
            device_code="",
            user_code="",
            verification_uri="",
            interval=0,
            expires_in=0,
        ),
        rebind=True,
    )
    token_response = ObjectProperty()

    def on_enter(self):
        self.start_device_flow()

    def start_device_flow(self):
        self.device_flow = auth.start_device_flow()
        print("Device flow started: ", self.device_flow)
        Clock.schedule_once(lambda dt: self.check_auth(), self.device_flow.interval)

    def open_browser(self, url):
        webbrowser.open(url)

    def check_auth(self):
        response = auth.ask_for_token(self.device_flow)
        match response.status:
            case auth.Status.AUTHORIZATION_PENDING:
                Clock.schedule_once(
                    lambda dt: self.check_auth(), self.device_flow.interval
                )
            case auth.Status.EXPIRED_TOKEN:
                self.start_device_flow()
            case auth.Status.ACCESS_DENIED:
                raise RuntimeError(
                    "TODO nice error message when user denies TreadI App"
                )
            case auth.Status.ACCESS_GRANTED:
                # Listeners on the token_response property are notified here
                self.token_response = response
            case _:
                raise RuntimeError(
                    f"TODO nice error message when other error encountered{response}"
                )


class TreadIApp(App):

    gql_client = None
    issue_loader = None
    issue_cache = IssueCache()
    sm = None

    def make_client_from_response(self, token_response):
        if token_response.status == auth.Status.ACCESS_GRANTED:
            self.gql_client = make_gql_client(token_response.access_token)
            return True
        return False

    def on_login_result(self, _, token_response):
        if self.make_client_from_response(token_response):
            auth.store_refresh_token(token_response.refresh_token)
            self.sm.transition.direction = "left"
            self.sm.current = "repos"
        else:
            raise RuntimeError(
                f"TODO more graceful response to login failure {token_response}"
            )

    def build(self):
        # Window.always_on_top = True

        self.sm = ScreenManager()

        token_response = auth.cycle_cached_token()
        if not self.make_client_from_response(token_response):
            # Ask user to login
            login_screen = LoginScreen()
            login_screen.bind(token_response=self.on_login_result)
            self.sm.add_widget(login_screen)

        self.sm.add_widget(RepoPickerScreen(name="repos"))
        self.sm.add_widget(IssueScreen(name="issues"))

        return self.sm


def main():
    TreadIApp().run()


if __name__ == "__main__":
    main()
